/*
 * Copyright (c) 2020 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include <string.h>
#include <zephyr/kernel.h>
#include <stdlib.h>
#include <zephyr/net/socket.h>
#include <zephyr/net/tls_credentials.h>
#include <zephyr/net/net_if.h>
#include <zephyr/net/wifi_mgmt.h>
#include <zephyr/net/net_event.h>

LOG_MODULE_REGISTER(app, LOG_LEVEL_INF);

//#include <net/wifi_credentials.h>

#define WIFI_SOCKET_MGMT_EVENTS (NET_EVENT_WIFI_CONNECT_RESULT | \
		    NET_EVENT_WIFI_DISCONNECT_RESULT)

#define HTTPS_PORT "443"

#define HTTPS_HOSTNAME "google.com"
//#define HTTPS_HOSTNAME "example.com"

#define HTTP_HEAD                                                                                  \
	"HEAD / HTTP/1.1\r\n"                                                                      \
	"Host: " HTTPS_HOSTNAME ":" HTTPS_PORT "\r\n"                                              \
	"Connection: close\r\n\r\n"

#define HTTP_HEAD_LEN (sizeof(HTTP_HEAD) - 1)

#define HTTP_HDR_END "\r\n\r\n"

#define RECV_BUF_SIZE 2048
#define TLS_SEC_TAG 42

static const char send_buf[] = HTTP_HEAD;
static char recv_buf[RECV_BUF_SIZE];

/* Certificate for `example.com` */
static const char cert[] = {
//#include "../cert/DigiCertGlobalRootCA.pem"
#include "../cert/globalsign.pem"
//#include "../cert/gtsr1.pem"
//#include "../cert/gsr4.pem"
//#include "../cert/cust.pem"
//#include "../cert/gts1c3.pem"
};

K_SEM_DEFINE(wifi_connected, 0, 1);
K_SEM_DEFINE(ipv4_obtained, 0, 1);

static bool wifi_is_connected;
static bool ip_obtained;

static struct net_mgmt_event_callback mgmt_cb;
static struct net_mgmt_event_callback net_mgmt_ipv4_callback;

static void wifi_mgmt_event_handler(struct net_mgmt_event_callback *cb,
			  uint32_t mgmt_event, struct net_if *iface)
{

    const struct wifi_status *wifi_status = (const struct wifi_status *)cb->info;

	switch (mgmt_event) {
	case NET_EVENT_WIFI_CONNECT_RESULT:
		if (wifi_status->status) {
			LOG_INF("Connection attempt failed, status code: %d", wifi_status->status);
			return;
		}
        LOG_INF("Wi-Fi Connected, waiting for IP address");
        wifi_is_connected = true;
		k_sem_give(&wifi_connected);
		break;
	case NET_EVENT_WIFI_DISCONNECT_RESULT:
		LOG_INF("Disconnected");
        wifi_is_connected = false;
		break;
	default:
        LOG_ERR("Unknown event: %d", mgmt_event);
		break;
	}
}

static void ipv4_mgmt_event_handler(struct net_mgmt_event_callback *cb,
				    uint32_t event, struct net_if *iface)
{
	switch (event) {
	case NET_EVENT_IPV4_ADDR_ADD:
		LOG_INF("IPv4 address acquired");
		ip_obtained = true;
		k_sem_give(&ipv4_obtained);
		break;
	case NET_EVENT_IPV4_ADDR_DEL:
		LOG_INF("IPv4 address lost");
		ip_obtained = false;
		break;
	default:
		LOG_DBG("Unknown event: 0x%08X", event);
		return;
	}
}

static int __wifi_args_to_params(struct wifi_connect_req_params *params)
{
	params->timeout = SYS_FOREVER_MS;

	/* SSID */
	params->ssid = CONFIG_WIFI_CREDENTIALS_STATIC_SSID;
	params->ssid_length = strlen(params->ssid);

#if defined(CONFIG_WIFI_CREDENTIALS_STATIC_TYPE_PSK)
	params->security = 1;
#elif defined(CONFIG_WIFI_CREDENTIALS_STATIC_TYPE_PSK_SHA256)
	params->security = 2;
#elif defined(CONFIG_WIFI_CREDENTIALS_STATIC_TYPE_SAE)
	params->security = 3;
#else
	params->security = 0;
#endif

#if !defined(CONFIG_WIFI_CREDENTIALS_STATIC_TYPE_NONE)
	params->psk = CONFIG_WIFI_CREDENTIALS_STATIC_PASSWORD;
	params->psk_length = strlen(params->psk);
#endif
	params->channel = WIFI_CHANNEL_ANY;

	/* MFP (optional) */
	params->mfp = WIFI_MFP_OPTIONAL;

	return 0;
}

int wifi_connect(void)
{
	struct net_if *iface = net_if_get_default();
    static struct wifi_connect_req_params cnx_params;

	__wifi_args_to_params(&cnx_params);

	if (iface == NULL) {
		LOG_ERR("Returned network interface is NULL");
		return -1;
	}

	int err = net_mgmt(NET_REQUEST_WIFI_CONNECT, iface,
    &cnx_params, sizeof(struct wifi_connect_req_params));

	if (err) {
		LOG_ERR("Connecting to Wi-Fi failed. error: %d", err);
		return ENOEXEC;
	}

    LOG_INF("Connection requested");
    return 0;
}

static void net_init(void)
{
     if (IS_ENABLED(CONFIG_NET_CONNECTION_MANAGER)) {
		LOG_INF("Registering wifi events");
		net_mgmt_init_event_callback(&mgmt_cb,
					     wifi_mgmt_event_handler, WIFI_SOCKET_MGMT_EVENTS);
		net_mgmt_add_event_callback(&mgmt_cb);
		LOG_INF("Registering IPv4 events");
		net_mgmt_init_event_callback(&net_mgmt_ipv4_callback, ipv4_mgmt_event_handler,
				     NET_EVENT_IPV4_ADDR_ADD | NET_EVENT_IPV4_ADDR_DEL);
		net_mgmt_add_event_callback(&net_mgmt_ipv4_callback);

	}
	wifi_connect();
	LOG_INF("Wait for Wi-fi connection");
	k_sem_take(&wifi_connected, K_FOREVER);
	LOG_INF("Wait for DHCP");
	k_sem_take(&ipv4_obtained, K_FOREVER);	
}

BUILD_ASSERT(sizeof(cert) < KB(16), "Certificate too large");

/* Setup TLS options on a given socket */
int tls_setup(int fd)
{
	int err;
	int verify;

	/* Security tag that we have provisioned the certificate with */
	const sec_tag_t tls_sec_tag[] = {
		TLS_SEC_TAG,
	};
	#if 1
	/* This is a hack... */
	static bool first;
	if (!first) {
		first = true;
		err = tls_credential_add(tls_sec_tag[0], TLS_CREDENTIAL_CA_CERTIFICATE, cert, sizeof(cert));
		if (err) {
			return err;
	}
	}
	#endif
	/* Set up TLS peer verification */
	enum {
		NONE = 0,
		OPTIONAL = 1,
		REQUIRED = 2,
	};

	verify = OPTIONAL;

	err = setsockopt(fd, SOL_TLS, TLS_PEER_VERIFY, &verify, sizeof(verify));
	if (err) {
		printk("Failed to setup peer verification, err %d\n", errno);
		return err;
	}

	/* Associate the socket with the security tag
	 * we have provisioned the certificate with.
	 */
	err = setsockopt(fd, SOL_TLS, TLS_SEC_TAG_LIST, tls_sec_tag, sizeof(tls_sec_tag));
	if (err) {
		printk("Failed to setup TLS sec tag, err %d\n", errno);
		return err;
	}

	err = setsockopt(fd, SOL_TLS, TLS_HOSTNAME, HTTPS_HOSTNAME, sizeof(HTTPS_HOSTNAME) - 1);
	if (err) {
		printk("Failed to setup TLS hostname, err %d\n", errno);
		return err;
	}
	return 0;
}

int main(void)
{
	int err;
	int fd[2] = {-1, -1};
	char *p;
	int bytes;
	size_t off;
	struct addrinfo *res;
	struct addrinfo hints = {
		.ai_flags = AI_NUMERICSERV, /* Let getaddrinfo() set port */
		.ai_socktype = SOCK_STREAM,
		.ai_family = AF_INET,
	};
	char peer_addr[INET6_ADDRSTRLEN];

	k_sleep(K_SECONDS(1));
	printk("HTTPS client sample started\n\r");
	net_init();

	printk("Looking up %s\n", HTTPS_HOSTNAME);
	err = getaddrinfo(HTTPS_HOSTNAME, HTTPS_PORT, &hints, &res);
	if (err) {
		printk("getaddrinfo() failed, err %d\n", errno);
		return 0;
	}

	inet_ntop(res->ai_family, &((struct sockaddr_in *)(res->ai_addr))->sin_addr, peer_addr,
		  INET6_ADDRSTRLEN);
	printk("Resolved %s (%s)\n", peer_addr, net_family2str(res->ai_family));

	for (int i = 0; i < 2; i++) {
		printk("FD: %d\n", i);
		fd[i] = socket(res->ai_family, SOCK_STREAM , IPPROTO_TLS_1_2);

		if (fd[i] == -1) {
			printk("Failed to open socket!\n");
			goto clean_up;
		}

		/* Setup TLS socket options */
		err = tls_setup(fd[i]);
		if (err) {
			printk("TLS Setup failed!\n");
			goto clean_up;
		}

		printk("Connecting to %s:%d\n", HTTPS_HOSTNAME,
			ntohs(((struct sockaddr_in *)(res->ai_addr))->sin_port));
		err = connect(fd[i], res->ai_addr, res->ai_addrlen);
		if (err) {
			printk("connect() failed, err: %s\n", strerror(errno));
			goto clean_up;
		}

		off = 0;
		do {
			bytes = send(fd[i], &send_buf[off], HTTP_HEAD_LEN - off, 0);
			if (bytes < 0) {
				printk("send() failed, err %d\n", errno);
				goto clean_up;
			}
			off += bytes;
		} while (off < HTTP_HEAD_LEN);

		printk("Sent %d bytes\n", off);

		off = 0;
		do {
			bytes = recv(fd[i], &recv_buf[off], RECV_BUF_SIZE - off, 0);
			if (bytes < 0) {
				printk("recv() failed, err %d\n", errno);
				goto clean_up;
			}
			off += bytes;
		} while (bytes != 0 /* peer closed connection */);

		printk("Received %d bytes\n", off);

		/* Make sure recv_buf is NULL terminated (for safe use with strstr) */
		if (off < sizeof(recv_buf)) {
			recv_buf[off] = '\0';
		} else {
			recv_buf[sizeof(recv_buf) - 1] = '\0';
		}

		/* Print HTTP response */
		p = strstr(recv_buf, "\r\n");
		if (p) {
			off = p - recv_buf;
			recv_buf[off + 1] = '\0';
			printk("\n>\t %s\n\n", recv_buf);
		}

		printk("Finished with socket fd=%d\n", fd[i]);
		k_msleep(5000);
	}

clean_up:
	printk("Clean up\n");
	freeaddrinfo(res);
	(void)close(fd[0]);
	(void)close(fd[1]);
	return 0;
}

